@@ -1,39 +1,37 @@
-import re
-from collections import defaultdict, namedtuple
-
-from packaging.version import Version
-import pkg_resources
-import sqlalchemy as sa
-from sqlalchemy import inspect
-from sqlalchemy.dialects.postgresql.base import (
-    PGCompiler, PGDDLCompiler, PGIdentifierPreparer, PGTypeCompiler,
-    PGExecutionContext, PGDialect
-)
-from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2
-from sqlalchemy.dialects.postgresql.psycopg2cffi import PGDialect_psycopg2cffi
-from sqlalchemy.engine.default import DefaultDialect
-from sqlalchemy.sql.type_api import TypeEngine
-from sqlalchemy.engine import reflection
-from sqlalchemy.ext.compiler import compiles
-from sqlalchemy.sql.expression import (
-    BinaryExpression, BooleanClauseList, Delete
-)
-from sqlalchemy.types import (
-    VARCHAR, NullType, SMALLINT, INTEGER, BIGINT,
-    DECIMAL, REAL, BOOLEAN, CHAR, DATE, TIMESTAMP)
-from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
-
-from .commands import (
-    CopyCommand, UnloadFromSelect, Format, Compression, Encoding,
-    CreateLibraryCommand, AlterTableAppendCommand, RefreshMaterializedView
-)
-from .ddl import (
-    CreateMaterializedView, DropMaterializedView, get_table_attributes
-)
 import importlib
 import json
+import re
+from collections import defaultdict, namedtuple
+from logging import getLogger
+
+import pkg_resources
+import sqlalchemy as sa
+from packaging.version import Version
+from sqlalchemy import inspect
+from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
+from sqlalchemy.dialects.postgresql.base import (PGCompiler, PGDDLCompiler,
+                                                 PGDialect, PGExecutionContext,
+                                                 PGIdentifierPreparer,
+                                                 PGTypeCompiler)
+from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2
+from sqlalchemy.dialects.postgresql.psycopg2cffi import PGDialect_psycopg2cffi
+from sqlalchemy.engine import reflection
+from sqlalchemy.engine.default import DefaultDialect
+from sqlalchemy.ext.compiler import compiles
+from sqlalchemy.sql.expression import (BinaryExpression, BooleanClauseList,
+                                       Delete)
+from sqlalchemy.sql.type_api import TypeEngine
+from sqlalchemy.types import (BIGINT, BOOLEAN, CHAR, DATE, DECIMAL, INTEGER,
+                              REAL, SMALLINT, TIMESTAMP, VARCHAR, NullType)
+
+from .commands import (AlterTableAppendCommand, Compression, CopyCommand,
+                       CreateLibraryCommand, Encoding, Format,
+                       RefreshMaterializedView, UnloadFromSelect)
+from .ddl import (CreateMaterializedView, DropMaterializedView,
+                  get_table_attributes)
 
 sa_version = Version(sa.__version__)
+logger = getLogger(__name__)
 
 try:
     import alembic
@@ -41,7 +39,6 @@ except ImportError:
     pass
 else:
     from alembic.ddl import postgresql
-
     from alembic.ddl.base import RenameTable
     compiles(RenameTable, 'redshift')(postgresql.visit_rename_table)
 
@@ -181,6 +178,117 @@ RESERVED_WORDS = set([
     "without",
 ])
 
+REFLECTION_SQL = """\
+    SELECT
+        n.nspname as "schema",
+        c.relname as "table_name",
+        att.attname as "name",
+        format_encoding(att.attencodingtype::integer) as "encode",
+        format_type(att.atttypid, att.atttypmod) as "type",
+        att.attisdistkey as "distkey",
+        att.attsortkeyord as "sortkey",
+        att.attnotnull as "notnull",
+        pg_catalog.col_description(att.attrelid, att.attnum)
+        as "comment",
+        adsrc,
+        attnum,
+        pg_catalog.format_type(att.atttypid, att.atttypmod),
+        pg_catalog.pg_get_expr(ad.adbin, ad.adrelid) AS DEFAULT,
+        n.oid as "schema_oid",
+        c.oid as "table_oid"
+    FROM pg_catalog.pg_class c
+    LEFT JOIN pg_catalog.pg_namespace n
+        ON n.oid = c.relnamespace
+    JOIN pg_catalog.pg_attribute att
+        ON att.attrelid = c.oid
+    LEFT JOIN pg_catalog.pg_attrdef ad
+        ON (att.attrelid, att.attnum) = (ad.adrelid, ad.adnum)
+    WHERE n.nspname !~ '^pg_'
+        AND att.attnum > 0
+        AND NOT att.attisdropped
+        {schema_clause} {table_clause}
+    UNION
+    SELECT
+        view_schema as "schema",
+        view_name as "table_name",
+        col_name as "name",
+        null as "encode",
+        col_type as "type",
+        null as "distkey",
+        0 as "sortkey",
+        null as "notnull",
+        null as "comment",
+        null as "adsrc",
+        null as "attnum",
+        col_type as "format_type",
+        null as "default",
+        null as "schema_oid",
+        null as "table_oid"
+    FROM pg_get_late_binding_view_cols() cols(
+        view_schema name,
+        view_name name,
+        col_name name,
+        col_type varchar,
+        col_num int)
+    WHERE 1 {schema_clause} {table_clause}
+    UNION
+    SELECT c.schemaname AS "schema",
+        c.tablename AS "table_name",
+        c.columnname AS "name",
+        null AS "encode",
+        -- Spectrum represents data types differently.
+        -- Standardize, so we can infer types.
+        CASE
+            WHEN c.external_type = 'int' THEN 'integer'
+            WHEN c.external_type = 'float' THEN 'real'
+            WHEN c.external_type = 'double' THEN 'double precision'
+            WHEN c.external_type = 'timestamp'
+            THEN 'timestamp without time zone'
+            WHEN c.external_type ilike 'varchar%'
+            THEN replace(c.external_type, 'varchar', 'character varying')
+            WHEN c.external_type ilike 'decimal%'
+            THEN replace(c.external_type, 'decimal', 'numeric')
+            ELSE
+            replace(
+            replace(
+                replace(c.external_type, 'decimal', 'numeric'),
+                'char', 'character'),
+            'varchar', 'character varying')
+            END
+            AS "type",
+        false AS "distkey",
+        0 AS "sortkey",
+        null AS "notnull",
+        null as "comment",
+        null AS "adsrc",
+        c.columnnum AS "attnum",
+        CASE
+            WHEN c.external_type = 'int' THEN 'integer'
+            WHEN c.external_type = 'float' THEN 'real'
+            WHEN c.external_type = 'double' THEN 'double precision'
+            WHEN c.external_type = 'timestamp'
+            THEN 'timestamp without time zone'
+            WHEN c.external_type ilike 'varchar%'
+            THEN replace(c.external_type, 'varchar', 'character varying')
+            WHEN c.external_type ilike 'decimal%'
+            THEN replace(c.external_type, 'decimal', 'numeric')
+            ELSE
+            replace(
+            replace(
+                replace(c.external_type, 'decimal', 'numeric'),
+                'char', 'character'),
+            'varchar', 'character varying')
+            END
+            AS "format_type",
+        null AS "default",
+        s.esoid AS "schema_oid",
+        null AS "table_oid"
+    FROM svv_external_columns c
+    JOIN svv_external_schemas s ON s.schemaname = c.schemaname
+    WHERE 1 {schema_clause} {table_clause}
+    ORDER BY "schema", "table_name", "attnum";
+    """
+
 
 class RedshiftTypeEngine(TypeEngine):
 
@@ -625,6 +733,84 @@ class RedshiftDialectMixin(DefaultDialect):
             columns.append(column_info)
         return columns
 
+    @reflection.cache
+    def has_table(self, connection, table_name, schema=None, **kw):
+        if not schema:
+            schema = inspect(connection).default_schema_name
+
+        info_cache = kw.get('info_cache')
+        table = self._get_all_relation_info(connection,
+                                            schema=schema,
+                                            table_name=table_name,
+                                            info_cache=info_cache)
+
+        return True if table else False
+
+    @reflection.cache
+    def get_check_constraints(self, connection, table_name, schema=None, **kw):
+        table_oid = self.get_table_oid(
+            connection, table_name, schema, info_cache=kw.get("info_cache")
+        )
+        table_oid = 'NULL' if not table_oid else table_oid
+
+        result = connection.execute("""
+                        SELECT
+                            cons.conname as name,
+                            pg_get_constraintdef(cons.oid) as src
+                        FROM
+                            pg_catalog.pg_constraint cons
+                        WHERE
+                            cons.conrelid = {} AND
+                            cons.contype = 'c'
+                        """.format(table_oid))
+        ret = []
+        for name, src in result:
+            # samples:
+            # "CHECK (((a > 1) AND (a < 5)))"
+            # "CHECK (((a = 1) OR ((a > 2) AND (a < 5))))"
+            # "CHECK (((a > 1) AND (a < 5))) NOT VALID"
+            # "CHECK (some_boolean_function(a))"
+            # "CHECK (((a\n < 1)\n OR\n (a\n >= 5))\n)"
+
+            m = re.match(
+                r"^CHECK *\((.+)\)( NOT VALID)?$", src, flags=re.DOTALL
+            )
+            if not m:
+                logger.warning(f"Could not parse CHECK constraint text: {src}")
+                sqltext = ""
+            else:
+                sqltext = re.compile(
+                    r"^[\s\n]*\((.+)\)[\s\n]*$", flags=re.DOTALL
+                ).sub(r"\1", m.group(1))
+            entry = {"name": name, "sqltext": sqltext}
+            if m and m.group(2):
+                entry["dialect_options"] = {"not_valid": True}
+
+            ret.append(entry)
+        return ret
+
+    @reflection.cache
+    def get_table_oid(self, connection, table_name, schema=None, **kw):
+        """Fetch the oid for schema.table_name.
+        Return null if not found (external table does not have table oid)"""
+        schema_clause = (
+            "AND schema = '{schema}'".format(schema=schema) if schema else ""
+        )
+
+        result = connection.execute("""
+                        SELECT table_id
+                        FROM
+                            SVV_TABLE_INFO
+                        WHERE 1
+                            {schema_clause}
+                            AND "table" = '{table_name}';
+                        """.format(
+                                schema_clause=schema_clause,
+                                table_name=table_name)
+                            )
+
+        return result.scalar()
+
     @reflection.cache
     def get_pk_constraint(self, connection, table_name, schema=None, **kw):
         """
@@ -789,6 +975,7 @@ class RedshiftDialectMixin(DefaultDialect):
             schema = default_schema
         info_cache = kw.get('info_cache')
         all_relations = self._get_all_relation_info(connection,
+                                                    schema=schema,
                                                     info_cache=info_cache)
         relation_names = []
         for key, relation in all_relations.items():
@@ -826,6 +1013,8 @@ class RedshiftDialectMixin(DefaultDialect):
                                schema=None, **kw):
         info_cache = kw.get('info_cache')
         all_relations = self._get_all_relation_info(connection,
+                                                    schema=schema,
+                                                    table_name=table_name,
                                                     info_cache=info_cache)
         key = RelationKey(table_name, schema, connection)
         if key not in all_relations.keys():
@@ -839,7 +1028,8 @@ class RedshiftDialectMixin(DefaultDialect):
         info_cache = kw.get('info_cache')
         all_schema_columns = self._get_schema_column_info(
             connection,
-            schema,
+            schema=schema,
+            table_name=table_name,
             info_cache=info_cache
         )
         key = RelationKey(table_name, schema, connection)
@@ -851,6 +1041,8 @@ class RedshiftDialectMixin(DefaultDialect):
                                   schema=None, **kw):
         info_cache = kw.get('info_cache')
         all_constraints = self._get_all_constraint_info(connection,
+                                                        schema=schema,
+                                                        table_name=table_name,
                                                         info_cache=info_cache)
         key = RelationKey(table_name, schema, connection)
         if key not in all_constraints.keys():
@@ -859,6 +1051,18 @@ class RedshiftDialectMixin(DefaultDialect):
 
     @reflection.cache
     def _get_all_relation_info(self, connection, **kw):
+        schema = kw.get('schema', None)
+        schema_clause = (
+            "AND schema = '{schema}'".format(schema=schema) if schema else ""
+        )
+
+        table_name = kw.get('table_name', None)
+        table_clause = (
+            "AND relname = '{table}'".format(
+                table=table_name
+            ) if table_name else ""
+        )
+
         result = connection.execute(sa.text("""
         SELECT
           c.relkind,
@@ -878,9 +1082,26 @@ class RedshiftDialectMixin(DefaultDialect):
              LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
              JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner
         WHERE c.relkind IN ('r', 'v', 'm', 'S', 'f')
-          AND n.nspname !~ '^pg_'
-        ORDER BY c.relkind, n.oid, n.nspname;
-        """))
+          AND n.nspname !~ '^pg_' {schema_clause} {table_clause}
+        UNION
+        SELECT
+            'r' AS "relkind",
+            s.esoid AS "schema_oid",
+            s.schemaname AS "schema",
+            null AS "rel_oid",
+            t.tablename AS "relname",
+            null AS "diststyle",
+            s.esowner AS "owner_id",
+            u.usename AS "owner_name",
+            null AS "view_definition",
+            null AS "privileges"
+        FROM
+            svv_external_tables t
+            JOIN svv_external_schemas s ON s.schemaname = t.schemaname
+            JOIN pg_catalog.pg_user u ON u.usesysid = s.esowner
+        where 1 {schema_clause} {table_clause}
+        ORDER BY "relkind", "schema_oid", "schema";
+        """.format(schema_clause=schema_clause, table_clause=table_clause)))
         relations = {}
         for rel in result:
             key = RelationKey(rel.relname, rel.schema, connection)
@@ -890,77 +1111,46 @@ class RedshiftDialectMixin(DefaultDialect):
     # We fetch column info an entire schema at a time to improve performance
     # when reflecting schema for multiple tables at once.
     @reflection.cache
-    def _get_schema_column_info(
-        self, connection, schema=None, **kw
-    ):
+    def _get_schema_column_info(self, connection, **kw):
+        schema = kw.get('schema', None)
         schema_clause = (
             "AND schema = '{schema}'".format(schema=schema) if schema else ""
         )
-        all_columns = defaultdict(list)
-        result = connection.execute(sa.text("""
-        SELECT
-          n.nspname as "schema",
-          c.relname as "table_name",
-          att.attname as "name",
-          format_encoding(att.attencodingtype::integer) as "encode",
-          format_type(att.atttypid, att.atttypmod) as "type",
-          att.attisdistkey as "distkey",
-          att.attsortkeyord as "sortkey",
-          att.attnotnull as "notnull",
-          pg_catalog.col_description(att.attrelid, att.attnum)
-            as "comment",
-          adsrc,
-          attnum,
-          pg_catalog.format_type(att.atttypid, att.atttypmod),
-          pg_catalog.pg_get_expr(ad.adbin, ad.adrelid) AS DEFAULT,
-          n.oid as "schema_oid",
-          c.oid as "table_oid"
-        FROM pg_catalog.pg_class c
-        LEFT JOIN pg_catalog.pg_namespace n
-          ON n.oid = c.relnamespace
-        JOIN pg_catalog.pg_attribute att
-          ON att.attrelid = c.oid
-        LEFT JOIN pg_catalog.pg_attrdef ad
-          ON (att.attrelid, att.attnum) = (ad.adrelid, ad.adnum)
-        WHERE n.nspname !~ '^pg_'
-          AND att.attnum > 0
-          AND NOT att.attisdropped
-          {schema_clause}
-        UNION
-        SELECT
-          view_schema as "schema",
-          view_name as "table_name",
-          col_name as "name",
-          null as "encode",
-          col_type as "type",
-          null as "distkey",
-          0 as "sortkey",
-          null as "notnull",
-          null as "comment",
-          null as "adsrc",
-          null as "attnum",
-          col_type as "format_type",
-          null as "default",
-          null as "schema_oid",
-          null as "table_oid"
-        FROM pg_get_late_binding_view_cols() cols(
-          view_schema name,
-          view_name name,
-          col_name name,
-          col_type varchar,
-          col_num int)
-        WHERE 1 {schema_clause}
-        ORDER BY "schema", "table_name", "attnum";
-        """.format(schema_clause=schema_clause))
+
+        table_name = kw.get('table_name', None)
+        table_clause = (
+            "AND table_name = '{table}'".format(
+                table=table_name
+            ) if table_name else ""
         )
-        for col in result:
-            key = RelationKey(col.table_name, col.schema, connection)
-            all_columns[key].append(col)
+
+        all_columns = defaultdict(list)
+        with connection.connect() as cc:
+            result = cc.execute(sa.text(REFLECTION_SQL.format(
+                schema_clause=schema_clause,
+                table_clause=table_clause
+            )))
+
+            for col in result:
+                key = RelationKey(col.table_name, col.schema, connection)
+                all_columns[key].append(col)
 
         return dict(all_columns)
 
     @reflection.cache
     def _get_all_constraint_info(self, connection, **kw):
+        schema = kw.get('schema', None)
+        schema_clause = (
+            "AND schema = '{schema}'".format(schema=schema) if schema else ""
+        )
+
+        table_name = kw.get('table_name', None)
+        table_clause = (
+            "AND table_name = '{table}'".format(
+                table=table_name
+            ) if table_name else ""
+        )
+
         result = connection.execute(sa.text("""
         SELECT
           n.nspname as "schema",
@@ -970,7 +1160,7 @@ class RedshiftDialectMixin(DefaultDialect):
           t.conkey,
           a.attnum,
           a.attname,
-          pg_catalog.pg_get_constraintdef(t.oid, true) as condef,
+          pg_catalog.pg_get_constraintdef(t.oid, true)::varchar(512) as condef,
           n.oid as "schema_oid",
           c.oid as "rel_oid"
         FROM pg_catalog.pg_class c
@@ -980,9 +1170,25 @@ class RedshiftDialectMixin(DefaultDialect):
           ON t.conrelid = c.oid
         JOIN pg_catalog.pg_attribute a
           ON t.conrelid = a.attrelid AND a.attnum = ANY(t.conkey)
-        WHERE n.nspname !~ '^pg_'
-        ORDER BY n.nspname, c.relname
-        """))
+        WHERE n.nspname !~ '^pg_' {schema_clause} {table_clause}
+        UNION
+        SELECT
+            s.schemaname AS "schema",
+            c.tablename AS "table_name",
+            'p' as "contype",
+            c.tablename || '_pkey' as "conname",
+            array[1::SMALLINT] as "conkey",
+            1 as "attnum",
+            c.columnname as "attname",
+            'PRIMARY KEY (' || c.columnname  || ')'::VARCHAR(512) as "condef",
+            s.esoid AS "schema_oid",
+            null AS "rel_oid"
+        FROM
+            svv_external_columns c
+            JOIN svv_external_schemas s ON s.schemaname = c.schemaname
+        where 1 {schema_clause} {table_clause}
+        ORDER BY "schema", "table_name"
+        """.format(schema_clause=schema_clause, table_clause=table_clause)))
         all_constraints = defaultdict(list)
         for con in result:
             key = RelationKey(con.table_name, con.schema, connection)
@@ -1163,8 +1369,8 @@ class RedshiftDialect_redshift_connector(RedshiftDialectMixin, PGDialect):
         fns = []
 
         def on_connect(conn):
-            from sqlalchemy.sql.elements import quoted_name
             from sqlalchemy import util
+            from sqlalchemy.sql.elements import quoted_name
             conn.py_types[quoted_name] = conn.py_types[util.text_type]
 
         fns.append(on_connect)
